#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import grp
import os
import pwd


def is_root():
    """Returns whether this script is run with user id 0 (root)"""
    return os.getuid() == 0

def get_uid_gid(uid_name='nobody', gid_name='nogroup'):
    """Returns uid and gid for the given username and groupname"""
    uid = pwd.getpwnam(uid_name).pw_uid
    gid = grp.getgrnam(gid_name).gr_gid
    return uid, gid



# -*- coding: utf-8 -*-

import configparser
import logging
import os
import socket
import textwrap


logger = logging.getLogger(__name__)

config_filename = '/etc/myprox/myprox.conf'


class Configuration():
    """Class for reading/writing the configuration file"""

    def __init__(self):
        """Object initialization"""
        self._config = None
        self.cache_ssl = None

    def exists(self):
        """Checks whether the config file exists"""
        return os.path.isfile(self.filename)
  
    def read_config(self):
        """Reads the config file"""
        try:
            logger.debug('Attempting to read config file [{0}]'.format(self.filename))
            cfg = configparser.ConfigParser()
            cfg.read(self.filename)
            self._config = cfg
        except Exception as e:
            logger.warning('Config file [{0}] could not be read [{1}], using defaults'.format(self.filename, str(e)))
            self._config = dict()

    def is_true(self, value):
        """Checks whether the given value evaluates to True"""
        value = str(value).lower()
        return value not in ['0', 'false']

    @property
    def filename(self):
        """Return the name of the config file (incl. path)"""
        return config_filename

    @property
    def config(self):
        """Return the config dictionary"""
        if self._config is None:
            self.read_config()
        return self._config

    @property
    def nodes(self):
        """Return a dictionary of nodes (config sections except 'myprox') with their captions"""
        nodes = dict()
        for section in self.config.sections():
            if section != 'myprox':
                nodes[section] = self.config[section].get('caption', '[unnamed node]')
        return nodes

    def get(self, itemname, default=None, node=None):
        """Return a specific item from the configuration or the provided default value if not present; section 'myprox' provides defaults"""
        value = None
        if node is not None:
            if node in self.nodes:
                value = self.config[node].get(itemname, None)
            else:
                logger.warn(f'Attempt to get config item [{itemname}] for an undefined node [{node}]')
        if value is None:
            value = self.config['myprox'].get(itemname, default)
        return value

    @property
    def sslcertfile(self):
        """The filename incl. path for the server certificate"""
        return os.path.join(os.path.dirname(self.filename), 'server.pem')

    @property
    def sslkeyfile(self):
        """The filename incl. path for the server private key"""
        return os.path.join(os.path.dirname(self.filename), 'key.pem')

    @property
    def socket_host(self):
        """The address to bind to"""
        return self.get('socket_host', '::')

    @property
    def socket_port(self):
        """The port to listen on"""
        return int(self.get('socket_port', 8080))

    @property
    def use_ssl(self):
        """Check whether we use SSL/TLS"""
        if self.cache_ssl is None:
            # We need to cache this because we might no longer have access permissions after dropping root privileges
            self.cache_ssl = os.path.exists(self.sslcertfile) and os.path.exists(self.sslkeyfile)
        return self.cache_ssl

    @property
    def url(self):
        """Get url of our webserver"""
        scheme = f'http{'s' if self.use_ssl else ''}'
        fqdn = socket.getfqdn()
        port = self.socket_port
        if (scheme == 'http') and (port == 80):
            port = ''
        if (scheme == 'https') and (port == 443):
            port = ''
        if port:
             port = ':' + str(port)
        return self.get('url', f'{scheme}://{fqdn}{port}')

    def proxmox_oidc_auth_domain(self, node):
        """The Proxmox domain used with OIDC authentication"""
        return self.get('proxmox_oidc_auth_domain', None, node)

    @property
    def oidc_redirect_url(self):
        """The MyProx callback URL for OIDC authentication"""
        return self.get('oidc_redirect_url', f'{self.url}/redirect_uri')

    @property
    def webserver_user(self):
        """In case MyProx is started as root, drop privileges to the given user for increased security"""
        return self.get('webserver_user', 'myprox')

    @property
    def webserver_group(self):
        """In case MyProx is started as root, drop privileges to the given group for increased security"""
        return self.get('webserver_group', 'myprox')

    @property
    def environment(self):
        """Defines the CherryPy runtime environment"""
        return self.get('environment', 'development')

    def proxmox_api(self, node):
        """The hostname/IP address of the Proxmox API"""
        return self.get('proxmox_api', 'localhost', node)

    def proxmox_api_endpoint(self, node):
        """Get Proxmox API base URL"""
        api_endpoint = f'https://{self.proxmox_api(node)}:8006/api2/json'
        return api_endpoint

    def proxmox_api_withport(self, node):
        """The hostname/IP address of the Proxmox API including port number"""
        result = self.proxmox_api(node)
        # Note: The following won't work with IPv6 addresses; we currently don't care since hostnames are usually used anyway
        if ':' not in result:
            result += ':8006'
        return result

    def proxmox_api_verifyssl(self, node):
        """Whether to check the ssl certificate of the Proxmox API"""
        return self.is_true(self.get('proxmox_api_verifyssl', 0, node))

    def proxmox_default_auth_domain(self, node):
        """Default authentication domain for the case that no domain is explicitly provided"""
        return self.get('proxmox_default_auth_domain', 'pve', node)
        
    def cookie_auth_domain(self, node):
        """The joint domain of MyProx and Proxmox to be set in an authentication cookie (used for authentication Proxmox' VNC client)"""
        domain = self.get('cookie_auth_domain', 'auto', node)
        if domain == 'auto':
            domain = self.proxmox_api # e.g. "myprox.prox.mydomain.com" or "192.168.1.0"
            parts = domain.rsplit('.', 2) # e.g. ['myprox.prox', 'mydomain', 'com'] or ['192.168', '1', '0']
            if not parts[-1].isnumeric(): # no IP address?
                domain = '.'.join(parts[-2:]) # e.g. "mydomain.com"
            # Don't set domain in case the default "localhost" is set
            if domain == 'localhost':
                domain = None
        return domain

    @property
    def login_caption(self):
        """Greeting text (in HTML format) to show on the login form"""
        default = '<h3 style="text-align: center; margin-bottom: 2em;">Welcome!</h3>'
        return self.get('login_caption', default)

    def shortcut_user(self, node):
        """The user to be used in case '.' is provided as username"""
        return self.get('shortcut_user', '', node)

    def shortcut_password(self, node):
        """The password to be used in case '.' is provided as username and no password given"""
        return self.get('shortcut_password', '', node)

    def machine_creation_url(self, node):
        """Email URL to be used for creating a machine"""
        return self.get('machine_creation_url', 'mailto:myadmin@mydomain.local?subject=New machine request&body=User: {username}', node)

    def expiry_prolongation_days(self, node):
        """Number of days from today when setting a new expiry date"""
        return int(self.get('expiry_prolongation_days', 365, node))

    def dryrun(self, node):
        """Define whether to disable sending any writing/changing requests to Proxmox API"""
        return self.is_true(self.get('dryrun', 0, node))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import cherrypy
import jinja2
import json
import logging
import os
import random
import requests
import string
import urllib.parse

import proxmoxer
from . import myproxapi
from . import setupenv


class WebApp():

    def __init__(self, cfg):
        """Instance initialization"""
        self.cfg = cfg
        self.jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')))

    @cherrypy.expose
    def index(self, action=None, id=None, action_selection=None):
        """Show a list of existing machines"""
        node = cherrypy.session.get('node')
        vms = cherrypy.session['proxmox'].get_virtual_machines()
        #cherrypy.log(str(vms), context='WEBAPP', severity=logging.INFO, traceback=False)
        tmpl = self.jinja_env.get_template('index.html')
        return tmpl.render(sessiondata=cherrypy.session, machines=vms)

    @cherrypy.expose
    def manage(self, action=None, id=None, action_selection=None):
        """Manage a machine"""
        node = cherrypy.session.get('node')
        machine_data = None
        message = None
        try:
            if id:
                action_results = {
                'start': 'start triggered',
                'reboot': 'reboot triggered',
                'shutdown': 'shutdown triggered',
                'reset': 'reset triggered', 
                'stop': 'stop (switch-off) triggered',
                'suspend': 'suspension (hibernation) triggered',
                'resume': 'resume triggered'
                }
                if action_selection in action_results.keys():
                    if not self.cfg.dryrun(node):
                        cherrypy.session['proxmox'].trigger_vm_action(id, action_selection)
                    message = f'Machine {action_results[action_selection]}.'
                elif action_selection == 'console':
                    raise cherrypy.HTTPRedirect(f'/console?id={id}')
                elif action_selection == 'console_vnc':
                    raise cherrypy.HTTPRedirect(f'/console_vnc?id={id}')
                elif action_selection == 'extend':
                    cherrypy.session['proxmox'].set_tag_expiry_bydays(id, self.cfg.expiry_prolongation_days(node))
                    message = 'The expiry data of this machine has been set according to the prolongation policy of your organization.'
                elif action_selection == 'destroy':
                    message = 'The functionality to destroy a machine is not yet implemented. . Contact support to do this.'
                elif action_selection == 'refresh':
                    pass # nothing to do
                elif action_selection is None: # no action triggered at all
                    pass # nothing to do
                else: # unknown action
                    message = 'Error: invalid action specified'
            else:
                message = 'Error: no identifier given'        
            if not machine_data and id:
                machine_data = cherrypy.session['proxmox'].get_virtual_machine_with_tags(id)
        except ValueError as e:
            message = f'\n{str(e)}'
        if machine_data is None:
            machine_data = dict()
        #cherrypy.log(str(machine_data), context='WEBAPP', severity=logging.WARNING, traceback=False)
        tmpl = self.jinja_env.get_template('manage.html')
        return tmpl.render(sessiondata=cherrypy.session, itemdata=machine_data, message=message)

    @cherrypy.expose
    def create(self, action=None, id=None):
        """Trigger creation of a VM"""
        node = cherrypy.session.get('node')
        raise cherrypy.HTTPRedirect(self.cfg.machine_creation_url(node).format(username=cherrypy.session['username']))

    @cherrypy.expose
    def console(self, id=None):
        """Provide a connection file for download"""
        cherrypy.log(f'Attempting to download connection file for [{id}] by user [{cherrypy.session["username"]}]', context='WEBAPP', severity=logging.INFO, traceback=False)
        try:
            file_dict = cherrypy.session['proxmox'].get_spice(id)
            file_data = '[virt-viewer]\n'
            for key, value in file_dict.items():
                file_data += f'{key}={value}\n'
            cherrypy.log(str(file_data), context='WEBAPP', severity=logging.DEBUG, traceback=False)
            filename = ''.join([random.choice(string.ascii_letters + string.digits) for n in range(6)])
            #cherrypy.response.headers['Content-Disposition'] = f'attachment; filename={filename}.vv'
            cherrypy.response.headers['Content-Disposition'] = f'inline; filename={filename}.vv'
            cherrypy.response.headers['Content-Type'] = 'application/x-virt-viewer'
            return file_data.encode('utf-8')
        except Exception as e:
            return str(e)

    @cherrypy.expose
    def console_vnc(self, id=None):
        """Open a VNC console for the provided VM using Proxmox' web console"""
        node = cherrypy.session.get('node')
        token = cherrypy.session['proxmox'].proxmox.get_tokens()[0]
        cherrypy.response.cookie['PVEAuthCookie'] = token        
        cherrypy.response.cookie['PVEAuthCookie']._coded_value = token # automatic encoding adds quotes; since these break Proxmox authentication, override automatic quoting
        cherrypy.response.cookie['PVEAuthCookie']['path'] = '/'
        domain = self.cfg.cookie_auth_domain(node)
        if domain is not None:
            cherrypy.response.cookie['PVEAuthCookie']['domain'] = domain
        cherrypy.response.cookie['PVEAuthCookie']['max-age'] = 60
        cherrypy.response.cookie['PVEAuthCookie']['samesite'] = 'Strict' # other values would be 'Lax' and 'None'
        if cherrypy.server.ssl_certificate is not None: # are we running on ssl/tls?
            cherrypy.response.cookie['PVEAuthCookie']['secure'] = 'true'
        cherrypy.response.cookie['PVEAuthCookie']['httponly'] = 'true'
        cherrypy.log(f'Auth cookie set: {cherrypy.response.cookie["PVEAuthCookie"].output()}', context='WEBAPP', severity=logging.DEBUG)
        # Proxmox does e.g. https://192.168.202.16:8006/?console=kvm&novnc=1&vmid=112&vmname=dh-testvm&node=dh-nas6&resize=off&cmd='
        prox = self.cfg.proxmox_api_withport(node)
        vmid, node = cherrypy.session['proxmox'].decompose_id(id)
        raise cherrypy.HTTPRedirect(f'https://{prox}/?console=kvm&novnc=1&vmid={vmid}&node={node}&resize=off&cmd=')

    @cherrypy.expose
    def start(self, id=None):
        """Trigger start of the provided VM"""
        raise cherrypy.HTTPRedirect('./manage?action_selection=start&' + urllib.parse.urlencode([('id', id)]))

    def get_myprox_instance(self, node, username, password=None, ticket=None):
        """Connect to ProxmoxAPI with provided credentials and store reference in session"""
        try:
            try:
                cherrypy.session['proxmox'] = myproxapi.MyProxAPI(self.cfg.proxmox_api(node), username, password, ticket, self.cfg.proxmox_api_verifyssl(node))
            except proxmoxer.backends.https.AuthenticationError as e:
                cherrypy.log(f'Wrong credentials for user ["{username}"]', context='WEBAPP', severity=logging.INFO, traceback=False)
                return 'invalid username/password'
        except Exception as e:
            cherrypy.log(f'Error accessing ProxmoxAPI by user ["{username}"]: {str(e)}', context='WEBAPP', severity=logging.WARNING, traceback=False)
            #raise # for debugging
            return 'Error accessing Proxmox - please try again later'
        return None

    def add_node_in_state_parameter(self, url, node):
        """Update the state query parameter of an URL to include node information in its JSON value"""
        # Parse the URL
        parsed_url = urllib.parse.urlparse(url)
        # Extract query parameters
        query_params = urllib.parse.parse_qs(parsed_url.query)
        # Decode the state parameter from URL encoding
        if 'state' not in query_params:
            raise ValueError("The 'state' parameter is not in the URL.")
        state_json = query_params['state'][0]
        state_dict = json.loads(state_json)
        # Add a node attribute to the state dictionary
        state_dict['node'] = node
        # Re-encode the modified state back to JSON
        updated_state_json = json.dumps(state_dict, separators=(',', ':'))
        # Update the query parameters with the new state
        query_params['state'] = updated_state_json
        # Reconstruct the URL
        updated_query = urllib.parse.urlencode(query_params, doseq=True)
        updated_url = urllib.parse.urlunparse(
            parsed_url._replace(query=updated_query)
        )
        return updated_url

    @cherrypy.expose
    def redirect_uri(self, code=None, state=None):
        """Handle the callback from the OIDC provider"""
        # Query parameter validation
        if (code is None) or (state is None):
            raise cherrypy.HTTPError(400, "Missing 'code' attribute or 'state' attribute in the response")
        # Extract node info from state
        state_dict = json.loads(state)
        node = state_dict.get('node')
        # Add a node attribute to the state dictionary
        state_dict['node']
        # Login request to Proxmox API
        api_endpoint = self.cfg.proxmox_api_endpoint(node) + '/access/openid/login'
        data = {
            'code': code,
            'state': state,
            'redirect-url': self.cfg.oidc_redirect_url
        }
        cherrypy.log(f'HTTP POST request to API at [{api_endpoint}] with data [{data}]', context='WEBAPP', severity=logging.DEBUG, traceback=False)
        response = requests.post(api_endpoint, data=data)
        if response.status_code == 200:
            result = response.json()
            username = result.get('data', {}).get('username')
            ticket = result.get('data', {}).get('ticket')
            if ticket and username:
                cherrypy.session['username'] = username
                # Connect to ProxmoxAPI with provided credentials and store reference in session
                error_text = self.get_myprox_instance(node, username, ticket=ticket)
                if error_text is not None:
                    raise cherrypy.HTTPError(500, error_text)
                raise cherrypy.HTTPRedirect('/')
            else:
                raise cherrypy.HTTPError(500, 'Authentication failed: Missing ticket or username')
        else:
            raise cherrypy.HTTPError(500, 'Failed to exchange auth code for Proxmox auth token')

    def trigger_oidc_auth(self, node):
        """Gets an auth URL from Proxmox and redirects to OIDC provider"""
        api_endpoint = self.cfg.proxmox_api_endpoint(node) + '/access/openid/auth-url'
        data = {
            'realm': self.cfg.proxmox_oidc_auth_domain(node),
            'redirect-url': self.cfg.oidc_redirect_url
        }
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        cherrypy.log(f'HTTP POST request to API at [{api_endpoint}] with redirect url [{self.cfg.oidc_redirect_url}]', context='WEBAPP', severity=logging.DEBUG, traceback=False)
        response = requests.post(api_endpoint, data=data, headers=headers)
        if response.status_code == 200:
            result = response.json()
            redirect_url = result.get('data')
            redirect_url = self.add_node_in_state_parameter(redirect_url, node)
            if redirect_url:
                raise cherrypy.HTTPRedirect(redirect_url)
            else:
                raise cherrypy.HTTPError(500, 'Failed to initiate OIDC authentication: invalid Proxmox API response')
        else:
            raise cherrypy.HTTPError(500, f'Failed to initiate OIDC authentication, status code [{response.status_code}]')

    def check_username_and_password(self, username, password):
        """Check whether provided username and password are valid when authenticating"""
        node = cherrypy.request.params.get('node')
        if node is not None:
            cherrypy.session['node'] = node
        # Shortcut user
        if username == '.':
            username = self.cfg.shortcut_user(node)
            if not len(password):
                password = self.cfg.shortcut_password(node)
        # OIDC login
        if (username == '') and self.cfg.proxmox_oidc_auth_domain(node):
            self.trigger_oidc_auth(node)
        # Add default domain in case no domain provided
        if '@' not in username:
            username += '@' + self.cfg.proxmox_default_auth_domain(node)
        # Connect to ProxmoxAPI with provided credentials and store reference in session
        error_text = self.get_myprox_instance(node, username, password)
        if error_text is not None:
            return error_text
        cherrypy.log(f'User ["{username}"] logged in', context='WEBAPP', severity=logging.INFO, traceback=False)
        return # credentials ok; all set

    def login_screen(self, from_page='..', username='', error_msg='', **kwargs):
        """Shows a login form"""
        nodes = self.cfg.nodes
        tmpl = self.jinja_env.get_template('login.html')
        login_caption = self.cfg.login_caption
        return tmpl.render(from_page=from_page, username=username, error_msg=error_msg, nodes=nodes, login_caption=login_caption).encode('utf-8')

    @cherrypy.expose
    def logout(self):
        """Ends the currently logged-in user's session"""
        username = cherrypy.session['username']
        cherrypy.session.clear()
        cherrypy.response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
        cherrypy.response.headers['Pragma'] = 'no-cache'
        cherrypy.response.headers['Expires'] = '0'
        raise cherrypy.HTTPRedirect('/', 302)
        return f'"{username}" has been logged out'

    def on_change_func(self):
        """React on config changes"""
        pass
        cherrypy.log('Error calling on_change_command', context='WEBAPP', severity=logging.ERROR, traceback=False)


def run_webapp(cfg):
    """Runs the CherryPy web application with the provided configuration data"""
    #logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    script_path = os.path.dirname(os.path.abspath(__file__))
    app = WebApp(cfg)
    # Use SSL/TLS if certificate files exist
    if cfg.use_ssl:
        cherrypy.server.ssl_module = 'builtin'
        cherrypy.server.ssl_certificate = cfg.sslcertfile
        cherrypy.server.ssl_private_key = cfg.sslkeyfile
    else:
        cherrypy.log(f'Not using SSL/TLS due to certificate files [{cfg.sslcertfile}] and [{cfg.sslkeyfile}] not being present', context='SETUP', severity=logging.WARNING, traceback=False)
    # Define socket parameters
    cherrypy.config.update({'server.socket_host': cfg.socket_host,
                            'server.socket_port': cfg.socket_port,
                           })
    # Disable autoreload (cannot listen at a port <1024 after dropping root privileges)
    cherrypy.config.update({'engine.autoreload.on': False})
    # Select environment
    if cfg.environment != 'development':
        cherrypy.config.update({'environment': cfg.environment})
    # Configure the web application
    app_conf = {
       '/': {
            'tools.sessions.on': True,
            'tools.sessions.secure': cfg.use_ssl,
            'tools.sessions.httponly': True,
            'tools.sessions.samesite': 'Strict',
            'tools.staticdir.root': os.path.join(script_path, 'webroot'),
            'tools.session_auth.on': True,
            'tools.session_auth.login_screen': app.login_screen,
            'tools.session_auth.check_username_and_password': app.check_username_and_password,
        },
        '/redirect_uri': {
            'tools.session_auth.on': False
        },
        '/static': {
            'tools.session_auth.on': False,
            'tools.staticdir.on': True,
            'tools.staticdir.dir': 'static'
        },
        '/favicon.ico': {
            'tools.session_auth.on': False,
            'tools.staticfile.on': True,
            'tools.staticfile.filename': os.path.join(script_path, 'webroot', 'static', 'favicon.ico')
        }
    }
    # Start CherryPy
    cherrypy.tree.mount(app, config=app_conf)
    if setupenv.is_root():
        # Drop privileges
        cherrypy.log(f'MyProx was started as root; attempting to drop privileges to user "{cfg.webserver_user}" and group "{cfg.webserver_group}"', context='SETUP', severity=logging.INFO, traceback=False)
        try:
            uid, gid = setupenv.get_uid_gid(cfg.webserver_user, cfg.webserver_group)
        except KeyError:
            cherrypy.log(f'Error dropping privileges to user "{cfg.webserver_user}" and group "{cfg.webserver_group}"; do they exist on your machine? Aborting.', context='SETUP', severity=logging.ERROR, traceback=False)
            exit(1)
        cherrypy.process.plugins.DropPrivileges(cherrypy.engine, umask=0o022, uid=uid, gid=gid).subscribe()
    cherrypy.engine.start()
    cherrypy.engine.signals.subscribe()
    cherrypy.engine.block()


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-

# Notes:
# - Tags: useful as meta information for e.g., provisioning or config management systems, see https://lists.proxmox.com/pipermail/pve-devel/2019-October/039967.html

import datetime

from . import proxapi


class MyProxAPI(proxapi.ProxAPI):

    def __init__(self, host, user, password=None, ticket=None, verify_ssl=True):
        """Instance initialization"""
        super().__init__(host, user, password=password, ticket=ticket, verify_ssl=verify_ssl)
        self.clear_tag_cache()

    def check_role_VMUserMyProx(self):
        """Check whether the role 'PVEVMUserMyProx' has the required permissions"""
        role = 'PVEVMUserMyProx'
        result = get_role(role)
        if not result:
            # See https://pve.proxmox.com/pve-docs/chapter-pveum.html#_privileges
            raise Exception(f'Proxmox role {role} does not yet exist. Please create it under "Datacenter"|"Permissions"|"Roles" and assign (at least) permissions "VM.Console", "VM.PowerMgmt", and "VM.Audit" to it.')
        if result.get('VM.PowerMgmt', 0) != 1: # needed to manage the status of the machine, e.g. start and stop it
            raise Exception(f'Proxmox role {role} needs to have "VM.PowerMgmt" permission assigned to it.')
        if result.get('VM.Console', 0) != 1: # needed to get access to the VM console
            raise Exception(f'Proxmox role {role} needs to have "VM.Console" permission assigned to it.')
        if result.get('VM.Audit', 0) != 1: # needed to read basic information
            raise Exception(f'Proxmox role {role} needs to have "VM.Audit" permission assigned to it.')

    def clear_tag_cache(self):
        """Clears the tag cache"""
        self._cache_id = None
        self._cache_tags = None

    def get_tags_direct(self, id):
        """Get a dictionary of all the tags assigned to a given virtual machine"""
        # Requires: ["perm","/vms/{vmid}",["VM.Audit"]]
        vmid, node = self.decompose_id(id)
        tags = self.proxmox.nodes(node).qemu(vmid).config.get().get('tags')
        if tags is None:
            tags = ''
        tags = tags.split(',')
        tags = [ tag.partition('.') for tag in tags ]
        tags = { key.strip(): value.strip() for key, _, value in tags }
        return tags        

    def get_tags(self, id):
        """Get a dictionary of all the tags assigned to a given virtual machine using tag cache"""
        if (self._cache_id is None) or (self._cache_tags is None) or (self._cache_id != id):
            self._cache_tags = self.get_tags_direct(id)
            self._cache_id = id
        return self._cache_tags 

    def set_tags(self, id, tags):
        """Overwrite the tags of a given virtual machine based on a dictionary of all the new tags"""
        # Requires permission: (/vms/{vmid}, VM.Config.Options)
        vmid, node = self.decompose_id(id)
        # Convert dict to string representation
        tags = [ {key} if (value is None) or (len(value) == 0) else f'{key}.{value}' for key, value in tags.items() ]
        tags = ', '.join(tags)
        self.proxmox.nodes(node).qemu(vmid).config.put(tags=tags)

    def ensure_tag_set(self, id, tag, value):
        """Make sure that the tags of a given virtual machine has the specified one set to a desired value"""
        tags = self.get_tags(id)
        tags[tag] = value
        return self.set_tags(id, tags)
        
    def ensure_tag_unset(self, id, tag):
        """Make sure that the tags of a given virtual machine don't include the specified one"""
        tags = self.get_tags(id)
        tags.pop(tag, None)
        return self.set_tags(id, tags)

    def get_tag_expiry(self, id):
        """Returns the value of the expiry tag"""
        tags = self.get_tags(id)
        expiry = tags.get('myprox_expiry')
        if expiry is not None:
            expiry = datetime.date.fromisoformat(expiry)
        return expiry

    def set_tag_expiry(self, id, newdate):
        """Sets the value of the expiry tag"""
        return self.ensure_tag_set(id, 'myprox_expiry', newdate.isoformat())

    def set_tag_expiry_bydays(self, id, days=365):
        """Sets the value of the expiry tag to the given number of days in the future"""
        newdate = datetime.date.today()+datetime.timedelta(days=days)
        return self.set_tag_expiry(id, newdate)

    def get_virtual_machine_with_tags(self, id):
        """Return the data of the given virtual machine (id format: 'vmid@node') incl. certain tags"""
        data = self.get_virtual_machine(id)
        if data is not None:
            self.clear_tag_cache()
            data['tag_expiry'] = self.get_tag_expiry(id)
        return data



import os
import setuptools


with open('README.md', 'r') as f:
    long_description = f.read()

setup_kwargs = {
    'name': 'MyProx',
    'version': '1.2.0',
    'author': 'Dirk Henrici',
    'author_email': 'pypi.myprox@henrici.name',
    'description': 'simple web frontend for Proxmox VE users',
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'url': 'https://www.github.com/towalink/myprox',
    'packages': setuptools.find_namespace_packages('src'),
    'package_dir': {'': 'src'},
    'include_package_data': True,
    'install_requires': ['cherrypy',
                         'requests',
                         'jinja2',
                         'proxmoxer',
                        ],
    'entry_points': '''
        [console_scripts]
        myprox=myprox:main
    ''',
    'classifiers': [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX :: Linux',
        #'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Topic :: System :: Systems Administration',
        'Framework :: CherryPy',
    ],
    'python_requires': '>=3.6',
    'keywords': 'Proxmox frontend gui',
    'project_urls': {
        'Project homepage': 'https://www.henrici.name/projects/myprox.html',
        'Repository': 'https://www.github.com/towalink/myprox',
        'Documentation': 'https://www.github.com/towalink/myprox',
    },
}


if __name__ == '__main__':
    setuptools.setup(**setup_kwargs)



#!/usr/bin/env python3

# ProxmoxAPI documentation: see https://pve.proxmox.com/pve-docs/api-viewer/

import importlib
import proxmoxer
from proxmoxer import SERVICES


def ProxmoxHTTPAuth_init(self, username, password, otp=None, base_url="", otptype="totp", ticket=None, **kwargs):
    """Patched ProxmoxHTTPAuth.__init__, see original at https://github.com/proxmoxer/proxmoxer/blob/develop/proxmoxer/backends/https.py"""
    # Replace super().__init__(...)
    backend_https = importlib.import_module(f'.backends.https', 'proxmoxer')
    backend_https.ProxmoxHTTPAuthBase.__init__(self, **kwargs)

    self.base_url = base_url
    self.username = username
    # Use ticket if available
    self.pve_auth_ticket = ticket if (ticket is not None) else ""

    self._get_new_tokens(password=password, otp=otp)

def Backend_init(
    self,
    host,
    user=None,
    password=None,
    ticket=None,
    otp=None,
    port=None,
    verify_ssl=True,
    mode="json",
    timeout=5,
    token_name=None,
    token_value=None,
    path_prefix=None,
    service="PVE",
    cert=None,
    proxies=None,
):
    """Patched Backends.__init__, see original file at https://github.com/proxmoxer/proxmoxer/blob/develop/proxmoxer/backends/https.py"""
    # Notes:
    # The only changes are:
    # - adding and passing on "ticket" keyword parameter
    # - allowing password==None if ticket is provided

    # Make relevant variables available
    backend_https = importlib.import_module(f'.backends.https', 'proxmoxer')
    ProxmoxHTTPAuth = backend_https.ProxmoxHTTPAuth
    ProxmoxHTTPApiTokenAuth = backend_https.ProxmoxHTTPApiTokenAuth
    config_failure = backend_https.config_failure

    self.proxies = proxies
    self.cert = cert
    host_port = ""
    if len(host.split(":")) > 2:  # IPv6
        if host.startswith("["):
            if "]:" in host:
                host, host_port = host.rsplit(":", 1)
        else:
            host = f"[{host}]"
    elif ":" in host:
        host, host_port = host.split(":")
    port = host_port if host_port.isdigit() else port

    # if a port is not specified, use the default port for this service
    if not port:
        port = SERVICES[service]["default_port"]

    self.mode = mode
    if path_prefix is not None:
        self.base_url = f"https://{host}:{port}/{path_prefix}/api2/{mode}"
    else:
        self.base_url = f"https://{host}:{port}/api2/{mode}"

    if token_name is not None:
        if "token" not in SERVICES[service]["supported_https_auths"]:
            config_failure("{} does not support API Token authentication", service)

        self.auth = ProxmoxHTTPApiTokenAuth(
            user,
            token_name,
            token_value,
            verify_ssl=verify_ssl,
            timeout=timeout,
            service=service,
            cert=self.cert,
            proxies=proxies,
        )
    # The following line got changed
    elif (password is not None) or (ticket is not None):
        if "password" not in SERVICES[service]["supported_https_auths"]:
            config_failure("{} does not support password authentication", service)

        self.auth = ProxmoxHTTPAuth(
            user,
            password,
            otp,
            base_url=self.base_url,
            # The following line got changed
            ticket=ticket,
            verify_ssl=verify_ssl,
            timeout=timeout,
            service=service,
            cert=self.cert,
            proxies=proxies,
        )
    else:
        config_failure("No valid authentication credentials were supplied")



class ProxAPI():

    def __init__(self, host, user, password=None, ticket=None, verify_ssl=True):
        """Object initialization: set API parameters"""
        # Monkey-patch "proxmoxer" library to support tickets instead of passwords
        backend = importlib.import_module(f'.backends.https', 'proxmoxer')
        backend.ProxmoxHTTPAuth.__init__ = ProxmoxHTTPAuth_init
        backend.Backend.__init__ = Backend_init
        # Create proxmoxer instance
        self.proxmox = proxmoxer.ProxmoxAPI(host, user=user, password=password, ticket=ticket, verify_ssl=verify_ssl)

    def int2human(self, value, decimal_places = -1):
        """Convert integer value to human readable one with 'K'/'M'/'G'/'T'"""
        letter = ''
        if value >= 1024:
            value /= 1024
            letter = 'K'
        if value >= 1024:
            value /= 1024
            letter = 'M'
        if value >= 1024:
            value /= 1024
            letter = 'G'
        if value >= 1024:
            value /= 1024
            letter = 'T'
        if decimal_places == -1: # automatically choose based on detail assumed to be needed
            decimal_places = 0 if value >= 10 else 1
        value = round(value, decimal_places)
        format = f'.{decimal_places}f'
        return f'{value:{format}}{letter}'
    
    def uptime2human(self, value, decimal_places = -1, lang = 'en'):
        """Convert uptime value to human readable one with 'seconds'/'minutes'/'hours'/'days'"""
        if value == 0:
            return 'n/a'
        unit = 'Sekunden' if lang == 'de' else 'seconds'
        if value >= 60:
            value /= 60
            unit = 'Minuten' if lang == 'de' else 'minutes'
        if value >= 60:
            value /= 60
            unit = 'Stunden' if lang == 'de' else 'hours'
            if value >= 24:
                value /= 24
                unit = 'Tage' if lang == 'de' else 'days'
        if decimal_places == -1: # automatically choose based on detail assumed to be needed
            decimal_places = 0 if value >= 10 else 1
        value = round(value, decimal_places)
        format = f'.{decimal_places}f'
        return f'{value:{format}} {unit}'
    
    def decompose_id(self, id):
        """Returns the vmid and node part of the provided id (format: 'vmid@node')"""
        if id is None:
            raise ValueError('No vmid provided')
        vmid, _, node = id.partition('@')
        if (node is not None) and (not len(node)):
            node = None
        if not vmid.isdecimal():
            raise ValueError('vmid must be numeric')
        return vmid, node
    
    def get_role(self, role):
        """Gets data regarding the specified role"""
        try:
            # all roles: proxmox.access.roles.get()
            # single role: proxmox.access.roles(role).get())
            result = self.proxmox.access.roles(role).get()
        except proxmoxer.core.ResourceException as e:
            err = str(e) # proxmoxer.core.ResourceException: 500 Internal Server Error: role 'MyRole' does not exist
            if 'does not exist' in err:
                return False
            raise
        return result
    
    def get_nodes(self):
        return self.proxmox.nodes.get()

    def get_virtual_machines(self, node=None, vmid=None, full=False):
        """Return the data of the available virtual machines (or filter to return data of VMs on single node or just the data of a single VM)"""
        result = dict()
        for current_node in ([{'node': node}] if node is not None else self.proxmox.nodes.get()):
            if current_node.get('status', 'online') != 'online': # one can't query non-online nodes
                continue
            # Works similarly for LXC:
            # for vm in self.proxmox.nodes(current_node["node"]).lxc.get():
            #     print("{0}: {1} => {2}".format(vm["vmid"], vm["name"], vm["status"]))
            #     print('Uptime:', uptime2human(vm['uptime']))
            for vm in self.proxmox.nodes(current_node['node']).qemu.get(full=(1 if full else 0)):
                item = vm.copy() #item = { key: value for key, value in vm.items() }
                item['node'] = current_node['node']
                item['mem_human'] = self.int2human(vm['mem'])
                item['maxmem_human'] = self.int2human(vm['maxmem'])
                if (item['mem_human'] == item['maxmem_human']) or (vm['mem'] == 0):
                    item['memrange'] = item['maxmem_human']
                else:
                    item['memrange'] = item['mem_human'] + ' of ' + item['maxmem_human']
                item['maxdisk_human'] = self.int2human(vm['maxdisk'])
                item['uptime_human'] = self.uptime2human(vm['uptime'])
                if item['status'] == 'running':
                    item['status_uptime'] = item['status'] + ' for ' + item['uptime_human']
                elif item['status'] == 'stopped':
                    item['status_uptime'] = item['status']
                else:
                    item['status_uptime'] = item['status'] + ', up for ' + item['uptime_human']
                result[vm['vmid']] = item
        if (vmid is not None):
            if str(vmid).isdecimal():
                return result.get(int(vmid))
            else:
                return None
        return result            

    def get_virtual_machine(self, id, full=False):
        """Return the data of the given virtual machine (id format: 'vmid@node')"""
        vmid, node = self.decompose_id(id)
        result = self.get_virtual_machines(node=node, vmid=vmid, full=full)
        return result

    def get_spice(self, id):
        """Gets the content of a SPICE connection file"""
        vmid, node = self.decompose_id(id)        
        try:
            result = self.proxmox.nodes(node).qemu(vmid).spiceproxy.post()
        except proxmoxer.core.ResourceException as e:
            err = str(e)
            if 'not running' in err:
                raise Exception(f'VM {vmid} on {node} not running [{err}]')
            elif 'no spice port' in err:
                raise Exception(f'VM {vmid} on {node} does not use SPICE [{err}]')
            else:
                raise
        return result

    def trigger_vm_action(self, id, action):
        """Triggers an action (status change) on the given VM"""
        vmid, node = self.decompose_id(id)
        if action == 'start':
            self.proxmox.nodes(node).qemu(vmid).status.start.post()
        elif action == 'reboot':
            self.proxmox.nodes(node).qemu(vmid).status.reboot.post()
        elif action == 'shutdown':
            self.proxmox.nodes(node).qemu(vmid).status.shutdown.post()
        elif action == 'reset':
            self.proxmox.nodes(node).qemu(vmid).status.reset.post()
        elif action == 'stop':
            self.proxmox.nodes(node).qemu(vmid).status.stop.post()
        elif action == 'suspend':
            self.proxmox.nodes(node).qemu(vmid).status.suspend.post()
        elif action == 'resume':
            self.proxmox.nodes(node).qemu(vmid).status.resume.post()
